package Fx_Maths_PTM;

// ---- Utilisation des librairies d'algèbre linéaire - début -----
// + Librairie Jama
import Jama.Matrix;
import ij.IJ;
import org.netlib.lapack.DGELSY;// fonction Scilab lsq (cas réel)
import org.netlib.lapack.DGGEV;// fonction Scilab spec (cas réel)
import org.netlib.util.intW;
/**
 * Fonctions utilitaires pour les librairies Jama et JLapack
 * @author VV
 */
public class CLinAlg {
	
	/**
	* Affiche la matrice reçue dans le log d'ImageJ
	* Le format affiché est copier-collable dans la console Scilab
	* @param m : une matrice Jama
	* @return  : rien
	*/
	static public void afficher_matrice_dans_log( Matrix m ) {
		int nNb_lignes = m.getRowDimension();
		int nNb_colonnes = m.getColumnDimension();
		for( int li = 0; li < nNb_lignes; ++li ) {
			String strLigne = "";
			if ( li == 0 ) {
				strLigne = "[";
			}
			for( int co = 0; co < nNb_colonnes; ++co ) {
				if ( co > 0 ) {
					strLigne += ", ";
				}
				strLigne += m.get(li, co);
				
			}
			if ( li < nNb_lignes-1 ) {
				strLigne += "; ";
			}
			else {
				strLigne += "]";
			}
			IJ.log(strLigne);
		}
	}// afficher_matrice_dans_log2
	
	
	/**
	 * Résoud au mieux le système A.x = b
	 * @param matA
	 * @param matB
	 * @param matX
	 * @return = 0: successful exit  valeur négative sinon
	 */
	static public int lsq( final Matrix matA, 
		                   final Matrix vB, 
		                   Matrix vX ) {
		return lsq( matA, vB, vX, 0.00001 );
	}// lsq
	
	
	/**
	 * Résoud au mieux le système A.x = b
	 * @param matA
	 * @param matB
	 * @param matX
	 * @param tol positive scalar, used to determine the effective rank of A 
	 *            (defined as the order of the largest leading triangular 
	 *             submatrix R11 in the QR factorization with pivoting of A, 
	 *             whose estimated condition number  The tol default value is set to sqrt(%eps).
	 * @return = 0: successful exit  valeur négative sinon
	 */
	static public int lsq( final Matrix matA, 
		                   final Matrix vB, 
		                   Matrix vX,
		                   double tol ) {
		// Vérification des pré-requis
		intW info = new intW( -1 );// calcul invalide
		if ( matA.getRowDimension() < matA.getColumnDimension() ) {
			return info.val;
		}
		if ( vB.getRowDimension() != matA.getRowDimension()) {
			return info.val;
		}
		// Appel de la fonction Lapack
		/*DGELSY ( int m,
				  int n,
				  int nrhs,
				  double[][] a,
				  double[][] b,
				  int[] jpvt,
				  double rcond,
				  intW rank,
				  double[] work,
				  int lwork,
				  intW info ) */
		// SUBROUTINE DGELSY( M, N, NRHS, A, LDA, B, LDB, JPVT, RCOND, RANK,
		//     $                   WORK, LWORK, INFO )
		/*  M       (input) INTEGER
		*          The number of rows of the matrix A.  M >= 0.
		*
		*  N       (input) INTEGER
		*          The number of columns of the matrix A.  N >= 0.
		*
		*  NRHS    (input) INTEGER
		*          The number of right hand sides, i.e., the number of
		*          columns of matrices B and X. NRHS >= 0.
		*
		*  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
		*          On entry, the M-by-N matrix A.
		*          On exit, A has been overwritten by details of its
		*          complete orthogonal factorization.
		*
		*  LDA     (input) INTEGER
		*          The leading dimension of the array A.  LDA >= max(1,M).
		*
		*  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
		*          On entry, the M-by-NRHS right hand side matrix B.
		*          On exit, the N-by-NRHS solution matrix X.
		*
		*  LDB     (input) INTEGER
		*          The leading dimension of the array B. LDB >= max(1,M,N).
		*
		*  JPVT    (input/output) INTEGER array, dimension (N)
		*          On entry, if JPVT(i) .ne. 0, the i-th column of A is permuted
		*          to the front of AP, otherwise column i is a free column.
		*          On exit, if JPVT(i) = k, then the i-th column of AP
		*          was the k-th column of A.
		*
		*  RCOND   (input) DOUBLE PRECISION
		*          RCOND is used to determine the effective rank of A, which
		*          is defined as the order of the largest leading triangular
		*          submatrix R11 in the QR factorization with pivoting of A,
		*          whose estimated condition number < 1/RCOND.
		*
		*  RANK    (output) INTEGER
		*          The effective rank of A, i.e., the order of the submatrix
		*          R11.  This is the same as the order of the submatrix T11
		*          in the complete orthogonal factorization of A.
		*
		*  WORK    (workspace/output) DOUBLE PRECISION array, dimension (MAX(1,LWORK))
		*          On exit, if INFO = 0, WORK(1) returns the optimal LWORK.
		*
		*  LWORK   (input) INTEGER
		*          The dimension of the array WORK.
		*          The unblocked strategy requires that:
		*             LWORK >= MAX( MN+3*N+1, 2*MN+NRHS ),
		*          where MN = min( M, N ).
		*          The block algorithm requires that:
		*             LWORK >= MAX( MN+2*N+NB*(N+1), 2*MN+NB*NRHS ),
		*          where NB is an upper bound on the blocksize returned
		*          by ILAENV for the routines DGEQP3, DTZRZF, STZRQF, DORMQR,
		*          and DORMRZ.
		*
		*          If LWORK = -1, then a workspace query is assumed; the routine
		*          only calculates the optimal size of the WORK array, returns
		*          this value as the first entry of the WORK array, and no error
		*          message related to LWORK is issued by XERBLA.
		*
		*  INFO    (output) INTEGER
		*          = 0: successful exit
		*          < 0: If INFO = -i, the i-th argument had an illegal value.*/
		// Préparation des parametres à passer à la fonction DGGEV
		int k;
		int nbLi = matA.getRowDimension();
		int nbCols = matA.getColumnDimension();
		int n = nbCols;

		// - copie des matrices Eigen dans des tableaux de données pour Lapack
		double[][] A = matA.getArrayCopy();

		// - copie des matrices Eigen dans des tableaux de données pour Lapack
		double[][] B = vB.getArrayCopy();
		double rcond = tol;
		/* public static void DGELSY(int m, 
									 int n, 
									 * int nrhs, 
									 * double[][] a, 
									 * double[][] b, 
									 * int[] jpvt, 
									 * double rcond, 
									 * intW rank, 
									 * double[] work, 
									 * int lwork, 
									 * intW info) */
		int nrhs = 1;// nb colonnes x et b
		int[] jpvt = new int[ nbCols ];
		intW rank = new intW( 0 );
		int lwork = Math.max(nbLi * nbCols + 3 * nbLi + 1, 
		                     2 * nbLi * nbCols + nbCols * nrhs );//MN+3*N+1, 2*MN+NRHS;
		double[] work = new double[ lwork ];
		DGELSY.DGELSY( nbLi,
		               nbCols,
		               nrhs,
		               A,
		               B,
		               jpvt,
		               rcond,
		               rank,
		               work,
		               lwork,
		               info );
		if ( info.val == 0 ){
			// Copie des résultats dans le vecteurs x
			for ( int i = 0; i < nbCols; ++i ) {
				vX.set( i, 0, B[i][0] );// résultat réél 
			}// for
		}
		return info.val;
	}// lsq
	
	
	/**
	* Résoud A.x = lambda.B.x avec des fonctions Lapack
	* lambda : valeurs propres
	*      x : vecteurs propres
	* Les matrices A et B doivent être carrées de même dimension
	* La matrice B peut-être non inversible
	* Dans cette fonction, je ne traite que le cas des solutions réelles
	* Utilise la fonction Lapack DGGEV
	* @param matA matrice A de A.x = lambda.B.x (non modifiée)
	* @param matB matrice B de A.x = lambda.B.x (non modifiée)
	* @param vValeurs_propres vecteur des valeurs propres (retourné)
	* @param matVecteurs_propres matrice qui contient les vecteurs propres (retourné)
	* @return info (si info == 0, pas d'erreur)
	*/
	static public int spec( final Matrix matA,
	                        final Matrix matB,
	                        Matrix vValeurs_propres,
	                        Matrix matVecteurs_propres ) {
		// Vérification des pré-requis
		intW info = new intW( -1 );// calcul invalide
		if ( matA.getRowDimension() != matA.getColumnDimension() ) {
			return info.val;
		}
		if ( matB.getRowDimension() != matB.getColumnDimension()) {
			return info.val;
		}
		if ( matA.getRowDimension() != matB.getRowDimension()) {
			return info.val;
		}

		/**  Definition:
		*  ===========
		*
		*       SUBROUTINE DGGEV( JOBVL, JOBVR, N, A, LDA, B, LDB, ALPHAR, ALPHAI,
		*                         BETA, VL, LDVL, VR, LDVR, WORK, LWORK, INFO )
		*
		*       .. Scalar Arguments ..
		*       CHARACTER          JOBVL, JOBVR
		*       INTEGER            INFO, LDA, LDB, LDVL, LDVR, LWORK, N
		*       ..
		*       .. Array Arguments ..
		*       DOUBLE PRECISION   A( LDA, * ), ALPHAI( * ), ALPHAR( * ),
		*      $                   B( LDB, * ), BETA( * ), VL( LDVL, * ),
		*      $                   VR( LDVR, * ), WORK( * )
		*       ....*/

		// Préparation des parametres à passer à la fonction DGGEV
		int k;
		int nbLi = matA.getRowDimension();
		int nbCols = matA.getColumnDimension();
		int n = nbCols;

		// - copie des matrices Eigen dans des tableaux de données pour Lapack
		double[][] A = matA.getArrayCopy();

		// - copie des matrices Eigen dans des tableaux de données pour Lapack
		double[][] B = matB.getArrayCopy();

		int lda = nbLi;
		int ldb = nbLi;
		int ldvl = nbLi;
		double[][] vl = new double[ ldvl ][ nbCols ];

		String jobvl = "V";
		String jobvr = "V";

		double[] vAlphar = new double[ nbLi ];
		double[] vAlphai = new double[ nbLi ];
		double[] vBeta = new double[ nbLi ];

		int ldvr = nbLi;
		double[][] vr = new double[ ldvr ][ nbCols ];

		int lwork = 8 * nbLi;// précision
		double[] work = new double[ lwork ];

		// Appel de la fonction Lapack
		/** public static DGGEV(
		 *			String jobvl, 
		 *			String jobvr, 
		 *			int n, 
		 *			double[][] a, 
		 *			double[][] b, 
		 *			double[] alphar, 
		 *			double[] alphai, 
		 *			double[] beta, 
		 *			double[][] vl, 
		 *			double[][] vr, 
		 *			double[] work, 
		 *			int lwork, 
		 *			org.netlib.util.intW info) */
		DGGEV.DGGEV( jobvl, 
		             jobvr, 
		             n, 
		             A, 
		             B, 
		             vAlphar, 
		             vAlphai, 
		             vBeta, 
		             vl, 
		             vr, 
		             work, 
		             lwork,
		             info );
		if ( info.val == 0 ){
			// Copie des résultats dans les matrices et vecteurs Eigen
			// - Calcul des valeurs propres
			for ( int i = 0; i < n; ++i ) {
				double beta = vBeta[i];
				if ( beta == 0.0 ) {
					beta = Math.exp( -100 );// pour éviter la division par zéro
				}
				vValeurs_propres.set( i, 0, vAlphar[i] / beta );// résultat réél ( vAlphai[i] doit être nul )
			}// for

			// - Calcul de la matrice des vecteurs propres
			k = 0;
			for ( int j = 0; j < nbCols; ++j ) {
				for ( int i = 0; i < nbLi; ++i ) {
					matVecteurs_propres.set( i, j, vl[ i ][ j ] );
					++k;
				}
			}// for
		}

		return info.val;
	}// spec
	
}// CLinAlg
